/*
    Copyright 2007 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include "bf2_sniff.h"
#include "show_dump2.h"
#include "rwbits.h"

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define READBITS    0
#define WRITEBITS   1
#define SEARCHSTART 0x00600000
#define SEARCHSIZE  0x00100000
#define READ1SEARCH "\x55\x8B\xEC\x83\xEC\x10\x8B\x55\x08\x85\xD2\x89\x4D\xF0\x75\x08" \
                    "\x32\xC0\x8B\xE5\x5D\xC2\x08\x00"
#define READ2SEARCH "\xd2\xe0\x8b\x4d\xf0\x2c\x01\x22\xd8\x88\x1a\x8b\x5d\x0c\x01\x59" \
                    "\x10\x5f\x5e\xb0\x01\x5b\x8b\xe5\x5d\xc2\x08\x00\xcc\xcc"
#define READ3SEARCH "\x55\x8b\xec\x83\xec\x10\x8b\x55\x0c\x85\xd2\x89\x4d\xf0\x75\x08" \
                    "\x32\xc0\x8b\xe5\x5d\xc2\x08\x00"
#define READ4SEARCH "\xd2\xe0\x2c\x01\x22\xd0\x88\x13\x8b\x55\x0c\x8b\x45\xf0\x01\x50" \
                    "\x10\x5f\x5b\xb0\x01\x5e\x8b\xe5\x5d\xc2\x08\x00\xcc\xcc"
#define WRITESEARCH "\x55\x8b\xec\x83\xec\x10\x8b\xd1\x8b\x4d\x0c\x85\xc9\x89\x55\xf0"



FILE        *fd = NULL;
void        *bf2_rbits1_off,
            *bf2_rbits2_off,
            *bf2_wbits_off;
static u8   dumpfile[64],
            bf2_rbits1_jmp[32],
            bf2_rbits2_jmp[32],
            bf2_rbits_stack[32],
            bf2_wbits_jmp[32];

typedef struct {
    u8      *packet;
    u32     packet_size;
    u32     zero1;
    u32     zero2;
    u32     rbits;
    u32     wbits;
} bfbits_t;



/*
Explanation of what I do

We have two functions to monito:
readbits which handles the incoming packets
writebits for the outgoing packets

While there are no problems for writebits since the content of the buffer is
already filled with the needed data we want to dump, there are problems with
readbits because we know where is located the buffer but when the function
is called naturally there is no data there, yet.
Then there is also another problem, the stack is corrupted when the function
ends because there are some works over the memory and so is not possible to
read the exact address of the buffer.

So in this case I have opted for a work-around that in my tests works
perfectly.
In short when the function is called I call a function which copy the stack
in a temporary memory location.
When the function ends I call the dumping function.

For avoiding redundant code I have opted for only one dumping function which
is able to recognize the read or write function automatically.
*/



void bf2_rbits_getargs(const void *args, ...) {
    memcpy(bf2_rbits_stack, (void *)&args, sizeof(void *) * 4);
}



void bf2_sniff(const void *args, ...) {
    bfbits_t    *bfbits;
    int     bits,
            left,
            len,
            rw;
    u8      *buff,
            *stack;
    char    show[150];

    stack  = (void *)&args;
    bfbits = *(bfbits_t **)(stack);
    rw = (bfbits->rbits) ? READBITS : WRITEBITS;
    if(rw == READBITS) {
        stack  = (void *)bf2_rbits_stack;
        bfbits = *(bfbits_t **)(stack);
    }
    buff   = *(u8 **)(stack + sizeof(void *) + sizeof(void *));
    bits   = *(int *)(stack + sizeof(void *) + sizeof(void *) + sizeof(void *));

    if(!fd) {
        fd = fopen((char *)dumpfile, "wb");
        if(!fd) exit(1);
        fputs(TITLE, fd);
        fprintf(fd,
            "------------------------------------------------------------------------------------------------------------------------------------\n"
            "%-66s%-66s\n"
            "------------------------------------------------------------------------------------------------------------------------------------\n",
            "RECV", "SEND");
    }

    left = (rw == READBITS) ? 0 : 66;

    if(
      ((rw == READBITS)  && !(bfbits->rbits - bits)) ||
      ((rw == WRITEBITS) && !bfbits->wbits)
    ) {
        fprintf(fd,
            "\n"
            "%*s==================================================================\n"
            "%*spacket size %08x\n",
            left, "", left, "", bfbits->packet_size);
    }

    len = sprintf(show, "%u", bits);
    memset(show, ' ', left);
    memset(show + left, '_', 66);
    sprintf(show + left + 66 - len, "%u\n", bits);
    if(bits <= 32) {
        fputs(show, fd);
        fprintf(fd, "%*s%08x\n", left, "", read_bits(bits, buff, 0));
    } else {
        memcpy(show + left + 31, "dump", 4);
        fputs(show, fd);
        show_dump(left, (void *)buff, ((bits + 7)&(~7)) >> 3, fd);
    }
    fflush(fd);
}



int jmp(u8 *buff, void *from, void *to) {
    buff[0] = 0xe9;
    *(u32 *)(buff + 1) = to - (from + 5);
    return(5);
}



int call(u8 *buff, void *from, void *to) {
    buff[0] = 0xe8;
    *(u32 *)(buff + 1) = to - (from + 5);
    return(5);
}



int mm(u8 *dst, u8 *src, int len) {
    memcpy(dst, src, len);
    return(len);
}



void *find_bytes(void *buff, int buffsz, u8 *data, int datalen) {
    u8      *p,
            *l;

    l = buff + buffsz - datalen;
    for(p = buff; p <= l; p++) {
        if(!memcmp(p, data, datalen)) return(p);
    }
    return(NULL);
}



void bf2_init(void) {
    u8      buff[512],
            *p;

    bf2_rbits1_off = find_bytes((void *)SEARCHSTART, SEARCHSIZE, (void *)READ1SEARCH, sizeof(READ1SEARCH) - 1);
    if(!bf2_rbits1_off) bf2_rbits1_off = find_bytes((void *)SEARCHSTART, SEARCHSIZE, (void *)READ3SEARCH, sizeof(READ3SEARCH) - 1);
    bf2_rbits2_off = find_bytes((void *)SEARCHSTART, SEARCHSIZE, (void *)READ2SEARCH, sizeof(READ2SEARCH) - 1);
    if(!bf2_rbits2_off) bf2_rbits2_off = find_bytes((void *)SEARCHSTART, SEARCHSIZE, (void *)READ4SEARCH, sizeof(READ4SEARCH) - 1);
    bf2_wbits_off  = find_bytes((void *)SEARCHSTART, SEARCHSIZE, (void *)WRITESEARCH, sizeof(WRITESEARCH) - 1);

    if(!bf2_rbits1_off || !bf2_rbits2_off || !bf2_wbits_off) {
        MessageBox(0,
            TITLE
            "Error: the readbits and writebits functions have not been found in the process\n"
            "If you are testing your client you need to use the no-cd executable since the original one is encrypted\n",
            HEAD,
            MB_OK | MB_TASKMODAL);
        exit(1);    }

        /**************\
        |* bf2_rbits1 *|
        \**************/

    p = buff;
    p += jmp(p, (void *)bf2_rbits1_off, (void *)bf2_rbits1_jmp);
    WriteProcessMemory(GetCurrentProcess(), (void *)bf2_rbits1_off, buff, p - buff, NULL);

    p = bf2_rbits1_jmp;
    *p++ = 0x51;                // push ecx
    p += call(p, p, (void *)bf2_rbits_getargs);
    *p++ = 0x59;                // pop ecx
    p += mm(p, (void *)"\x55\x8b\xec\x83\xec\x10", 6);  // restore the bytes
    p += jmp(p, (void *)p, (void *)bf2_rbits1_off + 6); // we have used previously

        /**************\
        |* bf2_rbits2 *|
        \**************/

    bf2_rbits2_off += (sizeof(READ2SEARCH) - 1) - 5;

    p = buff;
    p += jmp(p, (void *)bf2_rbits2_off, (void *)bf2_rbits2_jmp);
    WriteProcessMemory(GetCurrentProcess(), (void *)bf2_rbits2_off, buff, p - buff, NULL);

    p = bf2_rbits2_jmp;
    *p++ = 0x51;                // push ecx
    p += call(p, p, (void *)bf2_sniff);
    *p++ = 0x59;                // pop ecx
    p += mm(p, (void *)"\xb0\x01",     2);
    p += mm(p, (void *)"\xc2\x08\x00", 3);      // restore the bytes

        /*************\
        |* bf2_wbits *|
        \*************/

    p = buff;
    p += jmp(p, (void *)bf2_wbits_off, (void *)bf2_wbits_jmp);
    *p++ = 0x90;
    WriteProcessMemory(GetCurrentProcess(), (void *)bf2_wbits_off, buff, p - buff, NULL);

    p = bf2_wbits_jmp;
    *p++ = 0x51;                // push ecx
    p += call(p, p, (void *)bf2_sniff);
    *p++ = 0x59;                // pop ecx
    p += mm(p, (void *)"\x55\x8b\xec\x83\xec\x10", 6);  // restore the bytes
    p += jmp(p, (void *)p, (void *)bf2_wbits_off + 6);  // we have used previously

    sprintf((char *)dumpfile, "%08x_bf2_sniff.txt", (u32)time(NULL));

    sprintf((char *)buff,
        TITLE
        "DLL successfully injected and ready to work:\n"
        "- readbits  %p %p\n"
        "- writebits %p\n"
        "\n"
        "Dump file: %s\n"
        "\n"
        "Do you want to continue?\n",
        bf2_rbits1_off, bf2_rbits2_off,
        bf2_wbits_off,
        dumpfile);

    if(MessageBox(0,
        (char *)buff,
        HEAD,
        MB_OKCANCEL | MB_TASKMODAL) == IDCANCEL) exit(1);
}



BOOL APIENTRY DllMain(HMODULE hModule, DWORD ulReason, LPVOID lpReserved) {
    switch(ulReason) {
        case DLL_PROCESS_ATTACH: {
            DisableThreadLibraryCalls(hModule);
            bf2_init();
            } break;
        case DLL_THREAD_ATTACH: {
            } break;
        case DLL_PROCESS_DETACH: {
            if(fd) fclose(fd);
            } break;
        case DLL_THREAD_DETACH: {
            } break;
    }
    return(TRUE);
}


